@@ -2,7 +2,7 @@
  * test-bopomofo.c
  *
  * Copyright (c) 2012
- *      libchewing Core Team. See ChangeLog for details.
+ *      libchewing Core Team.
  *
  * See the file "COPYING" for information on usage and redistribution
  * of this file.
@@ -17,17 +17,10 @@
 #include <stdio.h>
 #include <string.h>
 
-#ifdef WITH_RUST
-#include "chewing_rs.h"
-#else
-#include "key2pho-private.h"
-#include "bopomofo-private.h"
-#endif
+#include "chewing.h"
 
-#include "plat_types.h"
 #include "testhelper.h"
-
-FILE *fd;
+#include "chewing-utf8-util.h"
 
 void test_select_candidate_no_rearward()
 {
@@ -53,7 +46,7 @@ void test_select_candidate_no_rearward()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_maxChiSymbolLen(ctx, 16);
 
@@ -62,10 +55,10 @@ void test_select_candidate_no_rearward()
     type_keystroke_by_string(ctx, "<D>");       /* ㄧˊㄕㄤˋㄌㄞˊ */
     ok_candidate(ctx, CAND_1, ARRAY_SIZE(CAND_1));
 
-    type_keystroke_by_string(ctx, "<D>");       /* ㄕㄤˋㄌㄞˊ */
+    type_keystroke_by_string(ctx, "<D>");       /* 移上 */
     ok_candidate(ctx, CAND_2, ARRAY_SIZE(CAND_2));
 
-    type_keystroke_by_string(ctx, "<D><D>2<E>");        /* select 移上來 */
+    type_keystroke_by_string(ctx, "<D><L><D>2<E>");        /* select 移上來 */
     ok_commit_buffer(ctx, CAND_1[1]);
 
     chewing_delete(ctx);
@@ -94,7 +87,7 @@ void test_select_candidate_rearward()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_maxChiSymbolLen(ctx, 16);
     chewing_set_phraseChoiceRearward(ctx, 1);
@@ -108,7 +101,7 @@ void test_select_candidate_rearward()
     type_keystroke_by_string(ctx, "<D>");       /* ㄕㄤˋㄌㄞˊ */
     ok_candidate(ctx, CAND_2, ARRAY_SIZE(CAND_2));
 
-    type_keystroke_by_string(ctx, "<D><D>2<E>");        /* select 移上來 */
+    type_keystroke_by_string(ctx, "<D><L><D>2<E>");        /* select 移上來 */
     ok_commit_buffer(ctx, CAND_1[1]);
 
     chewing_delete(ctx);
@@ -124,7 +117,7 @@ void test_select_candidate_no_rearward_with_symbol()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "hk4g4`31u6vu84" /* 測試，一下 */ );
 
@@ -186,7 +179,7 @@ void test_select_candidate_rearward_with_symbol()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_phraseChoiceRearward(ctx, 1);
 
     type_keystroke_by_string(ctx, "hk4g4`31u6vu84" /* 測試，一下 */ );
@@ -249,7 +242,7 @@ void test_select_candidate_no_rearward_start_with_symbol()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "`31hk4g4" /* ，測試 */ );
 
@@ -293,7 +286,7 @@ void test_select_candidate_rearward_start_with_symbol()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_phraseChoiceRearward(ctx, 1);
 
     type_keystroke_by_string(ctx, "`31hk4g4" /* ，測試 */ );
@@ -335,7 +328,7 @@ void test_del_bopomofo_as_mode_switch()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "2k");        /* ㄉㄜ */
     ok_bopomofo_buffer(ctx, "\xe3\x84\x89\xe3\x84\x9c" /* ㄉㄜ */ );
@@ -363,7 +356,7 @@ void test_select_candidate_4_bytes_utf8()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
     chewing_set_phraseChoiceRearward(ctx, 1);
     chewing_set_autoShiftCur(ctx, 1);
@@ -387,26 +380,28 @@ void test_select_candidate_in_middle_no_reaward()
 {
     ChewingContext *ctx;
     int ret;
-    const char *cand;
+    char *cand;
 
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "hk4g4u6<L><L>" /* 測試儀*/);
 
     ret = chewing_cand_open(ctx);
     ok(ret == 0, "chewing_cand_open return %d shall be %d", ret, 0);
 
-    cand = chewing_cand_string_by_index_static(ctx, 0);
+    cand = chewing_cand_string_by_index(ctx, 0);
     ok(strcmp(cand, "\xE9\x81\xA9\xE5\xAE\x9C") == 0, "first candidate `%s' shall be `%s'", cand, "\xE9\x81\xA9\xE5\xAE\x9C" /* 適宜 */);
+    chewing_free(cand);
 
     ret = chewing_cand_list_next(ctx);
     ok(ret == 0, "chewing_cand_list_next return %d shall be %d", ret, 0);
 
-    cand = chewing_cand_string_by_index_static(ctx, 0);
+    cand = chewing_cand_string_by_index(ctx, 0);
     ok(strcmp(cand, "\xE5\xB8\x82") == 0, "first candidate `%s' shall be `%s'", cand, "\xE5\xB8\x82" /* 市 */);
+    chewing_free(cand);
 
     chewing_delete(ctx);
 }
@@ -420,7 +415,7 @@ void test_select_candidate_in_middle_reaward()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_phraseChoiceRearward(ctx, 1);
 
     type_keystroke_by_string(ctx, "hk4g4u6<L><L>" /* 測試儀*/);
@@ -447,7 +442,7 @@ void test_select_candidate_second_page()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_candPerPage(ctx, 9);
     type_keystroke_by_string(ctx, "u4<D><R>4"); /* ㄧˋ */
@@ -468,12 +463,12 @@ void test_select_candidate_second_page_rewind()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_candPerPage(ctx, 9);
     chewing_set_spaceAsSelection(ctx, 1);
     chewing_set_phraseChoiceRearward(ctx, 1);
-    type_keystroke_by_string(ctx, "zp zp <D><D><R><D>"); /* ㄈㄣ ㄈㄣ */
+    type_keystroke_by_string(ctx, "zp zp <D><D><R><D><D>"); /* ㄈㄣ ㄈㄣ */
     ok_candidate(ctx, CAND, ARRAY_SIZE(CAND));
 
     chewing_delete(ctx);
@@ -500,7 +495,7 @@ void test_Esc_not_entering_chewing()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     type_keystroke_by_string(ctx, "<EE>");
     ok_keystroke_rtn(ctx, KEYSTROKE_IGNORE);
 
@@ -512,7 +507,7 @@ void test_Esc_in_select()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     type_keystroke_by_string(ctx, "`<EE>");
     ok_candidate(ctx, NULL, 0);
 
@@ -524,7 +519,7 @@ void test_Esc_entering_bopomofo()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     type_keystroke_by_string(ctx, "hk<EE>");
     ok_bopomofo_buffer(ctx, "");
 
@@ -536,7 +531,7 @@ void test_Esc_escCleanAllBuf()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_escCleanAllBuf(ctx, 1);
 
     type_keystroke_by_string(ctx, "hk4g4<EE>");
@@ -560,7 +555,7 @@ void test_Del_not_entering_chewing()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     type_keystroke_by_string(ctx, "<DC>");
     ok_keystroke_rtn(ctx, KEYSTROKE_IGNORE);
 
@@ -572,7 +567,7 @@ void test_Del_in_select()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     type_keystroke_by_string(ctx, "`<DC>");
     ok_keystroke_rtn(ctx, KEYSTROKE_ABSORB);    /* XXX: shall be ignore? */
 
@@ -584,11 +579,11 @@ void test_Del_word()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     type_keystroke_by_string(ctx, "hk4u g4<L><L><DC><E>");
-    ok_commit_buffer(ctx, "\xE6\xB8\xAC\xE8\xA9\xA6" /* 測試 */ );
+    ok_commit_buffer(ctx, "測試" );
 
     chewing_delete(ctx);
 }
@@ -605,7 +600,7 @@ void test_Backspace_not_entering_chewing()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     type_keystroke_by_string(ctx, "<B>");
     ok_keystroke_rtn(ctx, KEYSTROKE_IGNORE);
 
@@ -618,7 +613,7 @@ void test_Backspace_in_select()
     int ret;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "`<B>");
     ok_candidate(ctx, NULL, 0);
@@ -643,7 +638,7 @@ void test_Backspace_remove_bopomofo()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     type_keystroke_by_string(ctx, "hk<B>");
     ok_bopomofo_buffer(ctx, "\xE3\x84\x98" /* ㄘ */ );
 
@@ -655,11 +650,11 @@ void test_Backspace_word()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     type_keystroke_by_string(ctx, "hk4u g4<L><B><E>");
-    ok_commit_buffer(ctx, "\xE6\xB8\xAC\xE8\xA9\xA6" /* 測試 */ );
+    ok_commit_buffer(ctx, "測試" );
 
     chewing_delete(ctx);
 }
@@ -678,7 +673,7 @@ void test_Up_close_candidate_window_word()
     int ret;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "hk4");
     ret = chewing_cand_TotalChoice(ctx);
@@ -701,7 +696,7 @@ void test_Up_close_candidate_window_symbol()
     int ret;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "_");
     ret = chewing_cand_TotalChoice(ctx);
@@ -723,7 +718,7 @@ void test_Up_not_entering_chewing()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     type_keystroke_by_string(ctx, "<U>");
     ok_keystroke_rtn(ctx, KEYSTROKE_IGNORE);
 
@@ -743,7 +738,7 @@ void test_Down_open_candidate_window()
     int ret;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "hk4");
     ret = chewing_cand_TotalChoice(ctx);
@@ -766,7 +761,7 @@ void test_Down_reopen_symbol_candidate()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "_<D><R>");
     ok(chewing_cand_CurrentPage(ctx) == 1, "current page shall be 1");
@@ -782,7 +777,7 @@ void test_Down_not_entering_chewing()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     type_keystroke_by_string(ctx, "<D>");
     ok_keystroke_rtn(ctx, KEYSTROKE_IGNORE);
 
@@ -795,7 +790,7 @@ void test_Down_open_candidate_window_after_deleting_symbol()
     int ret;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "<<>hk4g4<<>" /* ，測試， */);
     ret = chewing_cand_TotalChoice(ctx);
@@ -826,7 +821,7 @@ void test_Tab_insert_breakpoint_between_word()
     IntervalType it;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     type_keystroke_by_string(ctx, "hk4g4<L>");
@@ -861,7 +856,7 @@ void test_Tab_connect_word()
     IntervalType it;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     type_keystroke_by_string(ctx, "u -4<L>");
@@ -895,7 +890,7 @@ void test_Tab_at_the_end()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "hk4g4u6vu84");
     ok_preedit_buffer(ctx, "\xE6\xB8\xAC\xE8\xA9\xA6\xE4\xB8\x80\xE4\xB8\x8B" /* 測試一下 */ );
@@ -927,7 +922,7 @@ void test_Capslock()
     int mode;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     mode = chewing_get_ChiEngMode(ctx);
     ok(mode == CHINESE_MODE, "mode shall be CHINESE_MODE");
@@ -942,9 +937,10 @@ void test_Capslock()
     mode = chewing_get_ChiEngMode(ctx);
     ok(mode == SYMBOL_MODE, "mode shall change to SYMBOL_MODE");
 
-    ok_bopomofo_buffer(ctx, "");
-    ok_preedit_buffer(ctx, "");
-    ok_commit_buffer(ctx, "");
+    type_keystroke_by_string(ctx, "<CB>");
+
+    mode = chewing_get_ChiEngMode(ctx);
+    ok(mode == CHINESE_MODE, "mode shall change to CHINESE_MODE");
 
     chewing_delete(ctx);
 }
@@ -955,7 +951,7 @@ void test_Home()
     int cursor;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     type_keystroke_by_string(ctx, "hk4g4");
@@ -975,7 +971,7 @@ void test_End()
     int cursor;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     type_keystroke_by_string(ctx, "hk4g4<L><L>");
@@ -995,7 +991,7 @@ void test_PageUp_not_entering_chewing()
     int cursor;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     type_keystroke_by_string(ctx, "hk4g4<L><L>");
@@ -1014,7 +1010,7 @@ void test_PageUp_in_select()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_candPerPage(ctx, 10);
 
@@ -1043,7 +1039,7 @@ void test_PageDown_not_entering_chewing()
     int cursor;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     type_keystroke_by_string(ctx, "hk4g4<L><L>");
@@ -1062,7 +1058,7 @@ void test_PageDown_in_select()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_candPerPage(ctx, 10);
 
@@ -1091,7 +1087,7 @@ void test_ShiftSpace()
     int mode;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     mode = chewing_get_ShapeMode(ctx);
     ok(mode == HALFSHAPE_MODE, "mode shall be HALFSHAPE_MODE");
@@ -1125,6 +1121,48 @@ void test_ShiftSpace()
     chewing_delete(ctx);
 }
 
+void test_ShiftSpaceDisabled()
+{
+    ChewingContext *ctx;
+    int mode;
+
+    ctx = chewing_new();
+    start_testcase(ctx);
+
+    chewing_config_set_int(ctx, "chewing.enable_fullwidth_toggle_key", 0);
+
+    mode = chewing_get_ShapeMode(ctx);
+    ok(mode == HALFSHAPE_MODE, "mode shall be HALFSHAPE_MODE");
+
+    type_keystroke_by_string(ctx, "<SS>");
+    mode = chewing_get_ShapeMode(ctx);
+    ok(mode == HALFSHAPE_MODE, "mode shall be HALFSHAPE_MODE");
+
+    type_keystroke_by_string(ctx, " ");
+    ok_commit_buffer(ctx, " "); /* Space */
+
+    chewing_set_ChiEngMode(ctx, SYMBOL_MODE);
+    type_keystroke_by_string(ctx, "a");
+    ok_commit_buffer(ctx, "a"); /* a */
+
+    chewing_set_ChiEngMode(ctx, CHINESE_MODE);
+    type_keystroke_by_string(ctx, "<SS>");
+    mode = chewing_get_ShapeMode(ctx);
+    ok(mode == HALFSHAPE_MODE, "mode shall be HALFSHAPE_MODE");
+
+    type_keystroke_by_string(ctx, " ");
+    ok_commit_buffer(ctx, " ");
+
+    type_keystroke_by_string(ctx, "hk4 <E>");
+    ok_commit_buffer(ctx, "冊 ");
+
+    chewing_set_ChiEngMode(ctx, SYMBOL_MODE);
+    type_keystroke_by_string(ctx, "a ");
+    ok_commit_buffer(ctx, " ");
+
+    chewing_delete(ctx);
+}
+
 void test_Numlock_numeric_input()
 {
     const TestData NUMLOCK_INPUT[] = {
@@ -1148,7 +1186,7 @@ void test_Numlock_numeric_input()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     for (i = 0; i < ARRAY_SIZE(NUMLOCK_INPUT); ++i) {
@@ -1177,7 +1215,7 @@ void test_Numlock_select_candidate()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     for (i = 0; i < ARRAY_SIZE(NUMLOCK_SELECT); ++i) {
@@ -1201,7 +1239,7 @@ void test_Space_empty_buffer()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_spaceAsSelection(ctx, 1);
 
     type_keystroke_by_string(ctx, " ");
@@ -1220,7 +1258,7 @@ void test_Space_selection_word()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_spaceAsSelection(ctx, 1);
 
     type_keystroke_by_string(ctx, "hk4g4<H>" /* 測試 */ );
@@ -1258,7 +1296,7 @@ void test_Space_selection_symbol()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_spaceAsSelection(ctx, 1);
 
     type_keystroke_by_string(ctx, "`");
@@ -1287,7 +1325,7 @@ void test_Space_selection_insert_eng_mode()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_spaceAsSelection(ctx, 1);
 
     type_keystroke_by_string(ctx, "hk4");
@@ -1306,6 +1344,117 @@ void test_Space()
     test_Space_selection_insert_eng_mode();
 }
 
+void test_FuzzySearchMode()
+{
+    const TestData FUZZY_INPUT[] = {
+        {"eji6aup6284cjo42941ul3<E>", "國民大會代表" },
+        {"eji aup 28 cjo 29 1ul <E>", "國民大會代表" },
+        {"ej au 2 cj 2 1 <E>", "國民大會代表" },
+        {"e a 2 c 2 1 <E>", "國民大會代表" },
+        {"ea2c21 <E>", "國民大會代表" },
+    };
+    size_t i;
+    ChewingContext *ctx;
+
+    ctx = chewing_new();
+    start_testcase(ctx);
+    chewing_set_maxChiSymbolLen(ctx, 16);
+    chewing_config_set_int(ctx, "chewing.conversion_engine", FUZZY_CHEWING_CONVERSION_ENGINE);
+
+    for (i = 0; i < ARRAY_SIZE(FUZZY_INPUT); ++i) {
+        type_keystroke_by_string(ctx, FUZZY_INPUT[i].token);
+        ok_commit_buffer(ctx, FUZZY_INPUT[i].expected);
+    }
+
+    chewing_delete(ctx);
+}
+
+void test_FuzzySearchMode_Hanyu()
+{
+    const TestData FUZZY_INPUT[] = {
+        {"guo2min2da4hui4dai4biao3<E>", "國民大會代表" },
+        {"guo min da hui dai biao <E>", "國民大會代表" },
+        {"g m d h d b <E>", "國民大會代表" },
+    };
+    size_t i;
+    ChewingContext *ctx;
+
+    ctx = chewing_new();
+    start_testcase(ctx);
+    chewing_set_maxChiSymbolLen(ctx, 16);
+    chewing_set_KBType(ctx, KB_HANYU_PINYIN);
+    chewing_config_set_int(ctx, "chewing.conversion_engine", FUZZY_CHEWING_CONVERSION_ENGINE);
+
+    for (i = 0; i < ARRAY_SIZE(FUZZY_INPUT); ++i) {
+        type_keystroke_by_string(ctx, FUZZY_INPUT[i].token);
+        ok_commit_buffer(ctx, FUZZY_INPUT[i].expected);
+    }
+
+    chewing_delete(ctx);
+}
+
+void test_SimpleEngine()
+{
+    const TestData SIMPLE_INPUT[] = {
+        {"ru031", "簡" },
+        {"20 1", "單" },
+        {"5j41", "住" },
+        {"up 1", "因" },
+        {"ai61", "模" },
+        {"g41", "市" },
+        {"!", "！" },
+        {"ru03<EE>", "" },
+        {"ru03<D>1", "筧" },
+        {"ru03<D><EE>", "" },
+    };
+    size_t i;
+    ChewingContext *ctx;
+
+    ctx = chewing_new();
+    start_testcase(ctx);
+    chewing_set_maxChiSymbolLen(ctx, 16);
+    chewing_config_set_int(ctx, "chewing.conversion_engine", SIMPLE_CONVERSION_ENGINE);
+
+    for (i = 0; i < ARRAY_SIZE(SIMPLE_INPUT); ++i) {
+        type_keystroke_by_string(ctx, SIMPLE_INPUT[i].token);
+        ok_commit_buffer(ctx, SIMPLE_INPUT[i].expected);
+    }
+
+    chewing_delete(ctx);
+}
+
+void test_Acknowledge()
+{
+    ChewingContext *ctx;
+
+    ctx = chewing_new();
+    start_testcase(ctx);
+
+    type_keystroke_by_string(ctx, "hk4g4<E>");
+    ok_commit_buffer(ctx, "測試");
+
+    chewing_ack(ctx);
+    ok_commit_buffer(ctx, "");
+
+    chewing_delete(ctx);
+}
+
+void test_BellCondition()
+{
+    ChewingContext *ctx;
+
+    ctx = chewing_new();
+    start_testcase(ctx);
+
+    type_keystroke_by_string(ctx, "zp zp<E>");
+    ok_preedit_buffer(ctx, "分");
+    ok_bopomofo_buffer(ctx, "ㄈㄣ");
+    ok_commit_buffer(ctx, "");
+    ok_keystroke_rtn(ctx, KEYSTROKE_ABSORB);
+
+    chewing_delete(ctx);
+}
+
 void test_get_phoneSeq()
 {
     static const struct {
@@ -1329,7 +1478,7 @@ void test_get_phoneSeq()
     unsigned short *phone;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 16);
 
     for (i = 0; i < ARRAY_SIZE(DATA); ++i) {
@@ -1355,7 +1504,7 @@ void test_bopomofo_buffer()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "1ul");
     ok_bopomofo_buffer(ctx, "\xE3\x84\x85\xE3\x84\xA7\xE3\x84\xA0" /* ㄅㄧㄠ */ );
@@ -1384,7 +1533,7 @@ void test_longest_phrase()
     IntervalType it;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "rup ji up6ji 1j4bj6y4ru32k7e.3ji "
                              /* ㄐㄧㄣ ㄨㄛ ㄧㄣˊ ㄨㄛ ㄅㄨˋ ㄖㄨˊ ㄗˋ ㄐㄧˇ ㄉㄜ˙ ㄍㄡˇ ㄨㄛ */
@@ -1407,7 +1556,7 @@ void test_auto_commit_phrase()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 3);
 
     type_keystroke_by_string(ctx, "hk4g4<L><T><L><D>1<EN>`31hk4" /* 測試，測 */ );
@@ -1426,7 +1575,7 @@ void test_auto_commit_symbol()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_maxChiSymbolLen(ctx, 2);
 
     type_keystroke_by_string(ctx, "`31hk4g4" /* ，測試 */ );
@@ -1448,7 +1597,7 @@ void test_interval()
     IntervalType it;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "`31hk4g4`31hk4g4`31" /* ，測試，測試， */ );
 
@@ -1478,7 +1627,7 @@ void test_jk_selection()
     const int EXPECT_CAND_LEN[] = { 1, 2, 1, 1, 2, 1, 1 };
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "`31hk4g4`31hk4g4`31" /* ，測試，測試， */ );
 
@@ -1507,7 +1656,7 @@ void test_KB_HSU()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_HSU);
 
@@ -1613,7 +1762,7 @@ void test_KB_HSU_example()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_HSU);
     chewing_set_phraseChoiceRearward(ctx, 1);
@@ -1654,11 +1803,11 @@ void test_KB_HSU_example()
     ok_preedit_buffer(ctx, "一隻隻可愛的小花貓");
     chewing_clean_preedit_buf(ctx);
 
-    type_keystroke_by_string(ctx, "sm sxajdwj<D><D>1xfsxajdgscewfhidxfdwj<D><D>1cd<D><D>1rnd");
+    type_keystroke_by_string(ctx, "sm sxajdwj<D><D>1xfsxajdgscewfhidxfdwj<D><D>1cd<D>1rnd");
     ok_preedit_buffer(ctx, "三歲到五歲的小孩五到十人");
     chewing_clean_preedit_buf(ctx);
 
-    type_keystroke_by_string(ctx, "jxl cewjdxl lxjxfdxmjleojcde cekj<D><D>1xfnxljcdve hwjjeoflod");
+    type_keystroke_by_string(ctx, "jxl cewjdxl lxjxfdxmjleojcde cekj<D><D>1xfnxljcdve hwj<D>1jeoflod");
     ok_preedit_buffer(ctx, "忠孝東路五段六十一巷五弄十七號九樓");
     chewing_clean_preedit_buf(ctx);
 
@@ -1706,7 +1855,7 @@ void test_KB_HSU_choice_append()
     const char *cand;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_KBType(ctx, KB_HSU);
 
     for (i = 0; i < ARRAY_SIZE(CHOICE_INFO_APPEND); ++i) {
@@ -1738,7 +1887,7 @@ void test_KB_HSU_choice_append_select()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_KBType(ctx, KB_HSU);
 
     type_keystroke_by_string(ctx, "k <D>4");
@@ -1769,7 +1918,7 @@ void test_KB_HSU_JVC()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_KBType(ctx, KB_HSU);
 
     for (int i = 0; i < ARRAY_SIZE(DATA); ++i) {
@@ -1796,13 +1945,32 @@ void test_KB_HSU_JVC()
     chewing_delete(ctx);
 }
 
+void test_KB_ET()
+{
+    ChewingContext *ctx;
+    ctx = chewing_new();
+    start_testcase(ctx);
+
+    chewing_set_KBType(ctx, KB_ET);
+    chewing_set_maxChiSymbolLen(ctx, 20);
+
+    type_keystroke_by_string(ctx, "hx3b8 p04mz ;y37u4'z3de4<E>");
+    ok_commit_buffer(ctx, "虎斑胖貓走去草地");
+    type_keystroke_by_string(ctx, "k84nez3=2;i4/x4/04tez4uw4j04ta <D>3ge84lew4ce9 ce3<E>");
+    ok_commit_buffer(ctx, "看鳥兒在樹上跳躍讓她見獵心喜");
+    type_keystroke_by_string(ctx, "vx- <D>9sxo ,r1/9 ;1ez4fq b9 .x 7u4<E>");
+    ok_commit_buffer(ctx, "弓縮著身子要飛奔出去");
+
+    chewing_delete(ctx);
+}
+
 void test_KB_ET26()
 {
     ChewingContext *ctx;
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_ET26);
 
@@ -1890,7 +2058,7 @@ void test_KB_ET26_choice_append()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
     chewing_set_KBType(ctx, KB_ET26);
 
     for (i = 0; i < ARRAY_SIZE(CHOICE_INFO_APPEND); ++i) {
@@ -1919,7 +2087,7 @@ void test_KB_DACHEN_CP26()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_DACHEN_CP26);
 
@@ -1989,13 +2157,50 @@ void test_KB_DACHEN_CP26()
     chewing_delete(ctx);
 }
 
+void test_KB_GIN_YIEH()
+{
+    ChewingContext *ctx;
+    ctx = chewing_new();
+    start_testcase(ctx);
+
+    chewing_set_KBType(ctx, KB_GIN_YIEH);
+    chewing_set_maxChiSymbolLen(ctx, 20);
+
+    type_keystroke_by_string(ctx, "v[a20 w;zsl u.ag'zjla3-z<E>");
+    ok_commit_buffer(ctx, "虎斑胖貓走去草地");
+    type_keystroke_by_string(ctx, "f0zd-la=qu9zh[zh;ze-lz',zn;ze8 <D>3t-0zc-,zb-p b-a<E>");
+    ok_commit_buffer(ctx, "看鳥兒在樹上跳躍讓她見獵心喜");
+    type_keystroke_by_string(ctx, "r[/ <D>9m[i 6k1hp u1-lzxo 2p y[ g'z<E>");
+    ok_commit_buffer(ctx, "弓縮著身子要飛奔出去");
+
+    chewing_delete(ctx);
+}
+
+void test_KB_IBM()
+{
+    ChewingContext *ctx;
+    ctx = chewing_new();
+    start_testcase(ctx);
+
+    chewing_set_KBType(ctx, KB_IBM);
+    chewing_set_maxChiSymbolLen(ctx, 20);
+
+    type_keystroke_by_string(ctx, "-s,1x 2v.3; iz,wd.o;,5a.<E>");
+    ok_commit_buffer(ctx, "虎斑胖貓走去草地");
+    type_keystroke_by_string(ctx, "0x.7a;,nmik.ys.yv.6a;.dj.uv.6f <D>3qax.8aj.eac ea,<E>");
+    ok_commit_buffer(ctx, "看鳥兒在樹上跳躍讓她見獵心喜");
+    type_keystroke_by_string(ctx, "9sb <D>9psg rh/yc i/a;.4l 1c ts wd.<E>");
+    ok_commit_buffer(ctx, "弓縮著身子要飛奔出去");
+
+    chewing_delete(ctx);
+}
 
 void test_KB_HANYU()
 {
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_HANYU_PINYIN);
 
@@ -2023,7 +2228,7 @@ void test_KB_HANYU_direct_symbol_output()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_HANYU_PINYIN);
 
@@ -2046,7 +2251,7 @@ void test_KB_THL()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_THL_PINYIN);
 
@@ -2075,7 +2280,7 @@ void test_KB_MPS2()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_MPS2_PINYIN);
 
@@ -2103,7 +2308,7 @@ void test_KB_DVORAK()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_DVORAK);
     type_keystroke_by_string(ctx, "kgl eh4gl 5l 2t7jl31s4");
@@ -2121,7 +2326,7 @@ void test_KB_DVORAK_HSU()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_DVORAK_HSU);
     type_keystroke_by_string(ctx, "idl vbcdl cl hu;jlynvc");
@@ -2134,12 +2339,28 @@ void test_KB_DVORAK_HSU()
     chewing_delete(ctx);
 }
 
+void test_KB_COLEMAK()
+{
+    ChewingContext *ctx;
+
+    ctx = chewing_new();
+    start_testcase(ctx);
+
+    chewing_set_KBType(ctx, KB_COLEMAK);
+    type_keystroke_by_string(ctx, "vl; sn4l; 5; 2e7c;31o4");
+    ok_preedit_buffer(ctx, "\xE6\x96\xB0\xE9\x85\xB7\xE9\x9F\xB3\xE7\x9C\x9F\xE7\x9A\x84\xE5\xBE\x88\xE6\xA3\x92"
+                      /* 新酷音真的很棒 */ );
+    chewing_clean_preedit_buf(ctx);
+
+    chewing_delete(ctx);
+}
+
 void test_KB_COLEMAK_DH_ANSI()
 {
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_COLEMAK_DH_ANSI);
     type_keystroke_by_string(ctx, "vl; sn4l; 5; 2e7d;31o4");
@@ -2156,7 +2377,7 @@ void test_KB_COLEMAK_DH_ORTH()
     ChewingContext *ctx;
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     chewing_set_KBType(ctx, KB_COLEMAK_DH_ORTH);
     type_keystroke_by_string(ctx, "dl; sn4l; 5; 2e7c;31o4");
@@ -2175,11 +2396,15 @@ void test_KB()
     test_KB_HSU_choice_append();
     test_KB_HSU_choice_append_select();
     test_KB_HSU_JVC();
+    test_KB_ET();
     test_KB_ET26();
     test_KB_ET26_choice_append();
     test_KB_DACHEN_CP26();
+    test_KB_GIN_YIEH();
+    test_KB_IBM();
     test_KB_DVORAK();
     test_KB_DVORAK_HSU();
+    test_KB_COLEMAK();
     test_KB_COLEMAK_DH_ANSI();
     test_KB_COLEMAK_DH_ORTH();
 
@@ -2211,7 +2436,7 @@ void test_chewing_phone_to_bopomofo()
      *  so the number for ㄆㄣ is (2<<9)+(0<<7)+(10<<3)+(0) = 1104
      */
 
-    start_testcase(NULL, fd);
+    start_testcase(NULL);
 
     u8phone = "\xE3\x84\x86\xE3\x84\xA3" /* ㄆㄣ */ ;
     phone = UintFromPhone(u8phone);
@@ -2264,7 +2489,7 @@ void test_static_buffer_reuse()
     clean_userphrase();
 
     ctx = chewing_new();
-    start_testcase(ctx, fd);
+    start_testcase(ctx);
 
     type_keystroke_by_string(ctx, "hk4g4ggg");
     ok_preedit_buffer(ctx, "測試");
@@ -2290,20 +2515,9 @@ void test_static_buffer_reuse()
 
 int main(int argc, char *argv[])
 {
-    char *logname;
-    int ret;
-
     putenv("CHEWING_PATH=" CHEWING_DATA_PREFIX);
     putenv("CHEWING_USER_PATH=" TEST_HASH_DIR);
 
-    ret = asprintf(&logname, "%s.log", argv[0]);
-    if (ret == -1)
-        return -1;
-    fd = fopen(logname, "w");
-    assert(fd);
-    free(logname);
-
-
     test_select_candidate();
     test_Esc();
     test_Del();
@@ -2318,8 +2532,14 @@ int main(int argc, char *argv[])
     test_PageUp();
     test_PageDown();
     test_ShiftSpace();
+    test_ShiftSpaceDisabled();
     test_Numlock();
     test_Space();
+    test_FuzzySearchMode();
+    test_FuzzySearchMode_Hanyu();
+    test_SimpleEngine();
+    test_Acknowledge();
+    test_BellCondition();
 
     test_get_phoneSeq();
     test_bopomofo_buffer();
@@ -2337,7 +2557,5 @@ int main(int argc, char *argv[])
 
     test_static_buffer_reuse();
 
-    fclose(fd);
-
     return exit_status();
 }
